module SurfaceGen
	class SurfaceGenTool
		def initialize()
			@active = false
			@cursor_id = []
			@cursor_id << UI.create_cursor(File.join(PLUGINS_DIR, SURFACEGEN_DIR, "cursor_surfacegen-0.png"), 0, 31)
			@cursor_id << UI.create_cursor(File.join(PLUGINS_DIR, SURFACEGEN_DIR, "cursor_surfacegen-1.png"), 0, 31)
			@cursor_id << UI.create_cursor(File.join(PLUGINS_DIR, SURFACEGEN_DIR, "cursor_surfacegen-2.png"), 0, 31)
			
			reset
		end
		
		def reset()
			@state = 0
			@ipold = Sketchup::InputPoint.new
			@ip = Sketchup::InputPoint.new
			@vec = Geom::Vector3d.new
			@dragging = false
			@p0 = nil
		end
	
		def activate()
			@active = true
			reset
		end
		
		def onCancel(flag, view)
			view.invalidate
			reset
		end
		
		def deactivate(view)
			@active = false
			view.invalidate
		end
		
		def suspend(view)
			@active = false
			view.invalidate
		end
		
		def resume(view)
			@active = true
			view.invalidate
		end
		
		def active?()
			@active
		end
		
		def onMouseMove(flags, x, y, view)
			@ignoreIp = false
			snapLength = Sketchup.active_model.options['UnitsOptions']['LengthSnapLength']
			
			case @state
			when 0
				@ip.pick view, x, y
				@p0 = @ip.position
				if @ip != @ipold then
					view.invalidate if (@ip.display? or @ipold.display?)
					
					@ipold.copy! @ip
					
					view.tooltip = @ip.tooltip
				end
				
				Sketchup::set_status_text "Dimensions", SB_VCB_LABEL
			when 1
				@ip.pick view, x, y, @ipold
				@p1 = @ip.position
				@p1.z = @p0.z
				
				canSnap = (@ip.degrees_of_freedom == 3) or (@ip.face and (@ip.face.normal.z.abs == 1.0))
				
				@offset = @p1 - @p0
				
				if canSnap then
					diff = (@offset.x.abs-@offset.y.abs).abs
					if (diff <= snapLength*25.0) then
						mag = ((@offset.x.abs+@offset.y.abs)*0.5/snapLength).round*snapLength
						@offset.x *= mag/@offset.x.abs
						@offset.y *= mag/@offset.y.abs
					end
					
					@offset.x = (@offset.x/snapLength).round*snapLength
					@offset.y = (@offset.y/snapLength).round*snapLength
					
					@p1 = @p0 + @offset
				end
				
				@p2 = @p0.clone
				@p3 = @p1.clone
				@p2[0], @p3[0] = @p3[0], @p2[0]
				
				Sketchup::set_status_text "Dimensions", SB_VCB_LABEL
				Sketchup::set_status_text Sketchup::format_length(@offset.x.abs) + ", " + Sketchup::format_length(@offset.y.abs), SB_VCB_VALUE
				
				view.invalidate
				
				if @offset.x.abs == @offset.y.abs then
					view.tooltip = "Square"
				else
					view.tooltip = @ip.tooltip
				end
				
				if ((x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
					@dragging = true
				end
			when 2
				@ip.pick view, x, y, @ipold
				@p4 = @ip.position
				
				screenPos = []
				screenPos << view.screen_coords(@p0 + @vec).x
				screenPos << view.screen_coords(@p1 + @vec).x
				screenPos << view.screen_coords(@p2 + @vec).x
				screenPos << view.screen_coords(@p3 + @vec).x
				
				distance = @p4.z - @p0.z
				@ignoreIp = ((@ip.degrees_of_freedom == 3) or (!@ip.valid?) or ((x >= screenPos.min-10) and (x <= screenPos.max+10)))
				if @ignoreIp then
					line1 = view.pickray(x, y)
					line2 = [@p1, Geom::Vector3d.new(0,0,1)]
					@p4 = Geom.closest_points(line1, line2)[0]
					distance = ((@p4.z - @p0.z)/snapLength).round*snapLength

					diff = (@offset.x.abs-distance.abs).abs
					if (diff <= snapLength*25.0) then
						distance *= @offset.x.abs/distance.abs
					end
					diff = (@offset.y.abs-distance.abs).abs
					if (diff <= snapLength*25.0) then
						distance *= @offset.y.abs/distance.abs
					end
				end
				
				@vec = Geom::Vector3d.new(0, 0, distance)
				
				Sketchup::set_status_text "Distance", SB_VCB_LABEL
				Sketchup::set_status_text Sketchup::format_length(distance), SB_VCB_VALUE
				
				view.invalidate
				
				if ((distance.abs == @offset.x.abs) or (distance.abs == @offset.y.abs)) then
					if (@offset.x.abs == @offset.y.abs) then
						view.tooltip = "Cube"
					else
						view.tooltip = "Cuboid"
					end
				else
					view.tooltip = @ignoreIp ? "" : @ip.tooltip
				end
			end
		end
		
		def advanceState()
			case @state
			when 0
				if @ip.valid? then
					@state = 1
				end
			when 1
				if @ip.valid? then
					@ipold.copy! @ip
					@state = 2
				end
			when 2
				if @ip.valid? then
					begin
						surfaceGen
					ensure
						reset
					end
				end
			end
		end
		
		def onLButtonDown(flags, x, y, view)
			if @state == 0 then
				@xdown = x
				@ydown = y
			end
		
			advanceState
		end
		
		def onLButtonUp(flags, x, y, view)
			if (@state == 1) and @dragging then
				advanceState
			end
		end
		
		def onReturn(view)
			advanceState
		end
		
		def onUserText(text, view)
			Sketchup::set_status_text "", SB_VCB_VALUE
			
			case @state
			when 1
				result = text.match(/^\s*([^,]+)\s*(?:,\s*([^,]+)\s*)?$/)
				return unless result
				
				sizex = result[1].to_l
				sizey = result[2] ? result[2].to_l : sizex
				return unless sizex and sizey
				
				@offset = @p1 - @p0
				@offset.x = (@offset.x < 0) ? -sizex : sizex
				@offset.y = (@offset.y < 0) ? -sizey : sizey
					
				@p1 = @p0 + @offset
				
				@p2 = @p0.clone
				@p3 = @p1.clone
				@p2[0], @p3[0] = @p3[0], @p2[0]
				
				view.invalidate
				
				advanceState
			when 2
				result = text.match(/^\s*([^,]+)\s*$/)
				return unless result
				
				size = result[1].to_l
				return unless size

				distance = ((@p4.z - @p0.z) < 0) ? -size : size
				
				@vec = Geom::Vector3d.new(0, 0, distance)
				
				view.invalidate
				
				advanceState
			end
		end
		
		def draw(view)
			if @state >= 1 then
				view.drawing_color = 0x000000
				view.line_width = 1
				
				if (@state == 1) and (@offset.x.abs == @offset.y.abs) then
					view.line_stipple = "-"
					view.draw_line @p0, @p1
				end
				
				view.line_stipple = ""
				view.draw_polyline @p0, @p2, @p1, @p3, @p0
				
				if @state >= 2 then
					view.draw_polyline @p0 + @vec, @p2 + @vec, @p1 + @vec, @p3 + @vec, @p0 + @vec
					view.draw_line @p0, @p0 + @vec
					view.draw_line @p1, @p1 + @vec
					view.draw_line @p2, @p2 + @vec
					view.draw_line @p3, @p3 + @vec	
				end
			end
			
			if @ip.display? and !@ignoreIp then
				@ip.draw view
			end
		end
		
		def getExtents()
			result = Geom::BoundingBox.new
			
			result.add @p0 if (@state >= 0) and @p0
			if (@state >= 1) and @p1 and @p2 and @p3 then
				result.add @p1
				result.add @p2
				result.add @p3
				if (@state >= 2) and @vec then
					result.add @p0 + @vec
					result.add @p1 + @vec
					result.add @p2 + @vec
					result.add @p3 + @vec
				end
			end
			
			return result
		end
		
		def onSetCursor()
			UI.set_cursor(@cursor_id[@state])
		end
		
		def enableVCB?()
			return @state != 0
		end
		
		def onKeyDown(key, repeat, flags, view)
			if (key == CONSTRAIN_MODIFIER_KEY) && (repeat == 1) then
				@shift_down_time = Time.now
				
				if view.inference_locked? then
					view.lock_inference
				elsif @ip.valid? then
					view.lock_inference @ip
				end
			end
		end

		def onKeyUp(key, repeat, flags, view)
			if (key == CONSTRAIN_MODIFIER_KEY) && view.inference_locked? && ((Time.now - @shift_down_time) > 0.5) then
				view.lock_inference
			end
		end
		
		def surfaceGen()
			begin
				Sketchup.active_model.start_operation "SurfaceGen Graph", true
			rescue
				Sketchup::active_model.start_operation "SurfaceGen Graph"
			end
		
			group = Sketchup::active_model.active_entities.add_group
			group.set_attribute(ATTR_DICTIONARY, "isGraph", true)
			group.set_attribute(ATTR_DICTIONARY, "graphType", "Function")
			group.set_attribute(ATTR_DICTIONARY, "eqn", "x^2+y^2-25")
			group.set_attribute(ATTR_DICTIONARY, "xcells", 25)
			group.set_attribute(ATTR_DICTIONARY, "ycells", 25)
			group.set_attribute(ATTR_DICTIONARY, "zcells", 25)
			group.set_attribute(ATTR_DICTIONARY, "xmin", -5)
			group.set_attribute(ATTR_DICTIONARY, "ymin", -5)
			group.set_attribute(ATTR_DICTIONARY, "zmin", -25)
			group.set_attribute(ATTR_DICTIONARY, "xmax", 5)
			group.set_attribute(ATTR_DICTIONARY, "ymax", 5)
			group.set_attribute(ATTR_DICTIONARY, "zmax", 25)
			group.set_attribute(ATTR_DICTIONARY, "softEdges", true)
			group.set_attribute(ATTR_DICTIONARY, "boundary", true)
			entities = group.entities
			
			SurfaceGen.add_boundary(entities)
			
			sizex = (@p1.x - @p0.x).abs
			sizey = (@p1.y - @p0.y).abs
			sizez = @vec.z.abs
			origin = Geom::Vector3d.new([@p0.x, @p1.x].min, [@p0.y, @p1.y].min, @p0.z + [0, @vec.z].min)
			
			group.transformation = Geom::Transformation.translation(origin)*Geom::Transformation.scaling(sizex/GRAPH_SIZE, sizey/GRAPH_SIZE, sizez/GRAPH_SIZE)
			group.add_observer SurfaceGenInstanceObserver.new
			
			Sketchup.active_model.selection.clear
			Sketchup.active_model.selection.add group
		
			Sketchup.active_model.commit_operation
		end
    end
end
