# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#   SurfaceGen plugin for Google SketchUp
# Created by Diggory Blake on 24 April 2011
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


require 'sketchup.rb'

module Math
  def self.max(a, b)
    a > b ? a : b
  end

  def self.min(a, b)
    a < b ? a : b
  end

  def self.clamp(a, b, c)
    (a < b ? b : (a > c ? c : a))
  end
  
  def self.abs(a)
    a.abs
  end
end

module SurfaceGen
	SURFACEGEN_DIR = "SURFACEGEN_Dir"
	GRAPH_SIZE = 1000.0
	WINDOW_HEIGHT = 320
	WINDOW_WIDTH = 250
	ATTR_DICTIONARY = "SurfaceGenAttr"
	PLATFORM = (Object::RUBY_PLATFORM =~ /mswin/i) ? :windows : ((Object::RUBY_PLATFORM =~ /darwin/i) ? :mac : :other)
	PLUGINS_DIR = File.dirname(__FILE__)
	
	def SurfaceGen.eqnToCode(eqn, name)
		eqn = eqn.to_s
		replacements = [
			[/\^/, "**"],
			[/([^a-z0-9:\.])sin(?=\()/, "\\1Math::sin"],
			[/([^a-z0-9:\.])cos(?=\()/, "\\1Math::cos"],
			[/([^a-z0-9:\.])tan(?=\()/, "\\1Math::tan"],
			[/([^a-z0-9:\.])asin(?=\()/, "\\1Math::asin"],
			[/([^a-z0-9:\.])acos(?=\()/, "\\1Math::acos"],
			[/([^a-z0-9:\.])atan(?=\()/, "\\1Math::atan"],
			[/([^a-z0-9:\.])sinh(?=\()/, "\\1Math::sinh"],
			[/([^a-z0-9:\.])cosh(?=\()/, "\\1Math::cosh"],
			[/([^a-z0-9:\.])tanh(?=\()/, "\\1Math::tanh"],
			[/([^a-z0-9:\.])asinh(?=\()/, "\\1Math::asinh"],
			[/([^a-z0-9:\.])acosh(?=\()/, "\\1Math::acosh"],
			[/([^a-z0-9:\.])atanh(?=\()/, "\\1Math::atanh"],
			[/([^a-z0-9:\.])atan2(?=\()/, "\\1Math::atan2"],
			[/([^a-z0-9:\.])sqrt(?=\()/, "\\1Math::sqrt"],
			[/([^a-z0-9:\.])cbrt(?=\()/, "\\1Math::cbrt"],
			[/([^a-z0-9:\.])exp(?=\()/, "\\1Math::exp"],
			[/([^a-z0-9:\.])hypot(?=\()/, "\\1Math::hypot"],
			[/([^a-z0-9:\.])log(?=\()/, "\\1Math::log"],
			[/([^a-z0-9:\.])ln(?=\()/, "\\1Math::log"],
			[/([^a-z0-9:\.])log2(?=\()/, "\\1Math::log2"],
			[/([^a-z0-9:\.])log10(?=\()/, "\\1Math::log10"],
			[/([^a-z0-9:\.])max(?=\()/, "\\1Math::max"],
			[/([^a-z0-9:\.])min(?=\()/, "\\1Math::min"],
			[/([^a-z0-9:\.])abs(?=\()/, "\\1Math::abs"],
			[/([^a-z0-9:\.])clamp(?=\()/, "\\1Math::clamp"],
			[/([^a-z0-9:\.])e(?=[^a-z0-9:\.])/, "\\1Math::E"],
			[/([^a-z0-9:\.])pi(?=[^a-z0-9:\.])/, "\\1Math::PI"]
		]
		
		eqn = " " + eqn + " "
		eqn.downcase!
		replacements.each do |r|
			eqn.gsub!(r[0], r[1])
		end
		
		begin
			x = 0.0
			y = 0.0
			z = 0.0
			eval(eqn)
		rescue Exception => exc
			UI.messagebox(name + ":\n"+exc.to_s)
			return nil
		end
		
		return eqn
	end
	
	def SurfaceGen.boolean(string)
		return true if string == true || string =~ /^true$/i
		return false if string == false || string.nil? || string =~ /^false$/i
		raise ArgumentError.new("invalid value for Boolean: \"#{string}\"")
	end

	def SurfaceGen.dialog()
		@dialog
	end
	
	def SurfaceGen.add_boundary(entities)
		componentDef = Sketchup::active_model.definitions["SurfaceGen_Boundaries"]
		if (!componentDef) then
			componentDef = Sketchup::active_model.definitions.add("SurfaceGen_Boundaries")
			boundaryEntities = componentDef.entities
		
			points = [
				[0, 0, 0], [GRAPH_SIZE, 0, 0], [0, GRAPH_SIZE, 0], [GRAPH_SIZE, GRAPH_SIZE, 0],
				[0, 0, GRAPH_SIZE], [GRAPH_SIZE, 0, GRAPH_SIZE], [0, GRAPH_SIZE, GRAPH_SIZE], [GRAPH_SIZE, GRAPH_SIZE, GRAPH_SIZE]
				]
				
			edges = []
			edges << boundaryEntities.add_line(points[0], points[1])
			edges << boundaryEntities.add_line(points[1], points[3])
			edges << boundaryEntities.add_line(points[3], points[2])
			edges << boundaryEntities.add_line(points[2], points[0])
			edges << boundaryEntities.add_line(points[4], points[5])
			edges << boundaryEntities.add_line(points[5], points[7])
			edges << boundaryEntities.add_line(points[7], points[6])
			edges << boundaryEntities.add_line(points[6], points[4])
			edges << boundaryEntities.add_line(points[0], points[4])
			edges << boundaryEntities.add_line(points[1], points[5])
			edges << boundaryEntities.add_line(points[2], points[6])
			edges << boundaryEntities.add_line(points[3], points[7])
			
			faces = []
			faces << boundaryEntities.add_face(points[0], points[1], points[3], points[2])
			faces << boundaryEntities.add_face(points[4], points[5], points[7], points[6])
			faces << boundaryEntities.add_face(points[0], points[1], points[5], points[4])
			faces << boundaryEntities.add_face(points[2], points[3], points[7], points[6])
			faces << boundaryEntities.add_face(points[0], points[2], points[6], points[4])
			faces << boundaryEntities.add_face(points[1], points[3], points[7], points[5])
			
			boundaryMaterial = Sketchup.active_model.materials["SurfaceGen_BoundaryFrontMaterial"]
			boundaryBackMaterial = Sketchup.active_model.materials["SurfaceGen_BoundaryBackMaterial"]
			
			if !boundaryMaterial then
				boundaryMaterial = Sketchup.active_model.materials.add "SurfaceGen_BoundaryFrontMaterial"
				boundaryMaterial.color = 0xFFFFFF
				boundaryMaterial.alpha = 0.5
			end
			if !boundaryBackMaterial then
				boundaryBackMaterial = Sketchup.active_model.materials.add "SurfaceGen_BoundaryBackMaterial"
				boundaryBackMaterial.color = 0x000000
				boundaryBackMaterial.alpha = 0.0
			end
			
			faces.each { |face|
				face.material = boundaryMaterial
				face.back_material = boundaryBackMaterial
				face.reverse!
			}
		end
		
		boundaryInstance = entities.add_instance(componentDef, Geom::Transformation.new)
		boundaryInstance.set_attribute(ATTR_DICTIONARY, "boundary", true)
		
		return boundaryInstance
	end
	
	def SurfaceGen.asyncLoop(range, &proc)
		from = range.min
		to = range.max
		iteration = lambda do |index|
			last = (index > to)
			
			proc.call(index, last)
			
			if !last then
				UI.start_timer(0.01, false) { iteration.call(index.succ) }
			end
		end
		
		iteration.call(from)
		
		return nil
	end
	
	def SurfaceGen.generate()
		begin
			Sketchup::active_model.start_operation "SurfaceGen", true
		rescue
			Sketchup::active_model.start_operation "SurfaceGen"
		end
	
		selection = Sketchup::active_model.selection
		return unless ((selection.length == 1) and (selection[0].get_attribute(ATTR_DICTIONARY, "isGraph", false)))
		
		surfaceGen = selection[0]
		
		if surfaceGen.get_attribute(ATTR_DICTIONARY, "graphType", "Function") == "Function" then
			SurfaceGen.generateFunction(surfaceGen)
		else
			SurfaceGen.generateEquation(surfaceGen)
		end
	end
	
	def SurfaceGen.finishGenerate(entities, mesh, softEdges, boundary)
		entities.clear!
		entities.fill_from_mesh mesh, false, softEdges ? 12 : 0
		
		if boundary then
			SurfaceGen.add_boundary(entities)
		end
		
		entities.model.commit_operation
	
		@dialog.execute_script("finished();");
	end
	
	def SurfaceGen.cancelGenerate(model)
		model.abort_operation
		
		@dialog.execute_script("finished();");
	end
	
	def SurfaceGen.generateFunction(surfaceGen)
		eqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "eqn", "x^2+y^2-25"), "Function")
		xmineqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "xmin", -5), "min. X")
		ymineqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "ymin", -5), "min. Y")
		zmineqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "zmin", -25), "min. Z")
		xmaxeqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "xmax", 5), "max. X")
		ymaxeqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "ymax", 5), "max. Y")
		zmaxeqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "zmax", 25), "max. Z")
		return SurfaceGen.cancelGenerate(surfaceGen.model) unless eqn and xmineqn and ymineqn and zmineqn and xmaxeqn and ymaxeqn and zmaxeqn
		
		xsize = surfaceGen.get_attribute(ATTR_DICTIONARY, "xcells", 25).to_i
		ysize = surfaceGen.get_attribute(ATTR_DICTIONARY, "ycells", 25).to_i
		
		if (xsize < 1) or (ysize < 1) or (xsize > 250) or (ysize > 250) then
			UI.messagebox("Invalid number of cells!")
			return SurfaceGen.cancelGenerate(surfaceGen.model)
		end
		
		begin
			xoffset = eval(xmineqn).to_f
			yoffset = eval(ymineqn).to_f
			zoffset = eval(zmineqn).to_f
			xscale = (eval(xmaxeqn).to_f - xoffset)/xsize
			yscale = (eval(ymaxeqn).to_f - yoffset)/ysize
			zscale = (eval(zmaxeqn).to_f - zoffset)
		rescue Exception => exc
			UI.messagebox("Exception while evaluating axis ranges!")
			return SurfaceGen.cancelGenerate(surfaceGen.model)
		end
		
		softEdges = boolean(surfaceGen.get_attribute(ATTR_DICTIONARY, "softEdges", true))
		boundary = boolean(surfaceGen.get_attribute(ATTR_DICTIONARY, "boundary", true))
		
		
		generateHeights(eqn, xsize, ysize, xoffset, xscale, yoffset, yscale) do |data|
			entities = surfaceGen.entities
			
			mesh = Geom::PolygonMesh.new
			
			points = Array2D.new(data.dx, data.dy)
			
			xmul = GRAPH_SIZE/xsize
			ymul = GRAPH_SIZE/ysize
			zmul = GRAPH_SIZE/zscale
			SurfaceGen.asyncLoop(0...(data.dy)) do |iy, last|
				if !last then
					(0...(data.dx)).each do |ix|
						next unless data[ix, iy]
						
						x = ix*xmul
						y = iy*ymul
						z = (data[ix, iy]-zoffset)*zmul
						
						next if z < 0
						next if z > GRAPH_SIZE
						
						points[ix, iy] = mesh.add_point(Geom::Point3d.new(x,y,z))
						
					end
				else
					SurfaceGen.asyncLoop(0...(data.dy-1)) do |iy, last|
						if !last then
							(0...(data.dx-1)).each do |ix|
								p1 = points[ix, iy]
								p2 = points[ix+1, iy]
								p3 = points[ix, iy+1]
								p4 = points[ix+1, iy+1]
								
								next unless (p1 and p2 and p3 and p4)
								
								mesh.add_polygon p1, p2, p4
								mesh.add_polygon p4, p3, p1
							end
						else
							SurfaceGen.finishGenerate(entities, mesh, softEdges, boundary)
						end
					end
				end
			end
		end
	end
	
	def SurfaceGen.generateEquation(surfaceGen)
		eqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "eqn", "x^2+y^2-25"), "Equation")
		xmineqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "xmin", -5), "min. X")
		ymineqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "ymin", -5), "min. Y")
		zmineqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "zmin", -25), "min. Z")
		xmaxeqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "xmax", 5), "max. X")
		ymaxeqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "ymax", 5), "max. Y")
		zmaxeqn = SurfaceGen.eqnToCode(surfaceGen.get_attribute(ATTR_DICTIONARY, "zmax", 25), "max. Z")
		return SurfaceGen.cancelGenerate(surfaceGen.model) unless eqn and xmineqn and ymineqn and zmineqn and xmaxeqn and ymaxeqn and zmaxeqn

		xsize = surfaceGen.get_attribute(ATTR_DICTIONARY, "xcells", 25).to_i
		ysize = surfaceGen.get_attribute(ATTR_DICTIONARY, "ycells", 25).to_i
		zsize = surfaceGen.get_attribute(ATTR_DICTIONARY, "zcells", 25).to_i
		
		if (xsize < 1) or (ysize < 1) or (zsize < 1) or (xsize > 250) or (ysize > 250) or (zsize > 250) then
			UI.messagebox("Invalid number of cells!")
			return SurfaceGen.cancelGenerate(surfaceGen.model)
		end
		
		begin
			xoffset = eval(xmineqn).to_f
			yoffset = eval(ymineqn).to_f
			zoffset = eval(zmineqn).to_f
			xscale = (eval(xmaxeqn).to_f - xoffset)/xsize
			yscale = (eval(ymaxeqn).to_f - yoffset)/ysize
			zscale = (eval(zmaxeqn).to_f - zoffset)/zsize
		rescue Exception => exc
			UI.messagebox("Exception while evaluating axis ranges!")
			return SurfaceGen.cancelGenerate(surfaceGen.model)
		end

		softEdges = boolean(surfaceGen.get_attribute(ATTR_DICTIONARY, "softEdges", true))
		boundary = boolean(surfaceGen.get_attribute(ATTR_DICTIONARY, "boundary", true))
		
		generateVolume(eqn, xsize+1, ysize+1, zsize+1, xoffset, xscale, yoffset, yscale, zoffset, zscale) do |data|
			entities = surfaceGen.entities
			
			mesh = Geom::PolygonMesh.new
			
			points = Array3D.new(data.dx, data.dy, data.dz)
			
			adjData = [0, 0, 0]
			
			xmul = GRAPH_SIZE/xsize
			ymul = GRAPH_SIZE/ysize
			zmul = GRAPH_SIZE/zsize
			muls = [xmul, ymul, zmul]
			
			SurfaceGen.asyncLoop(0...(data.dz)) do |iz, last|
				if !last then
					(0...(data.dy)).each do |iy|
						(0...(data.dx)).each do |ix|
							curData = data[ix, iy, iz]
							solid = (curData < 0)
							
							adjData[0] = (ix < (data.dx-1)) ? data[ix+1, iy, iz] : nil
							adjData[1] = (iy < (data.dy-1)) ? data[ix, iy+1, iz] : nil
							adjData[2] = (iz < (data.dz-1)) ? data[ix, iy, iz+1] : nil
							
							x = ix*xmul
							y = iy*ymul
							z = iz*zmul
							
							points[ix, iy, iz] = [nil, nil, nil]
							
							(0..2).each do |dirIndex|
								next unless adjData[dirIndex]
							
								adjSolid = (adjData[dirIndex] < 0)
								if (solid != adjSolid) then
									p = Geom::Point3d.new(x,y,z)
									p[dirIndex] += (curData/(adjData[dirIndex]-curData)).abs*muls[dirIndex]
									points[ix, iy, iz][dirIndex] = mesh.add_point(p)
								end
							end
						end
					end
				else
					edges = Array.new(12)
					SurfaceGen.asyncLoop(0...(data.dz-1)) do |iz, last|
						if !last then
							(0...(data.dy-1)).each do |iy|
								(0...(data.dx-1)).each do |ix|
									cubeIndex = 0
									cubeIndex += 1 if (data[ix, iy, iz] < 0)
									cubeIndex += 2 if (data[ix+1, iy, iz] < 0)
									cubeIndex += 4 if (data[ix+1, iy+1, iz] < 0)
									cubeIndex += 8 if (data[ix, iy+1, iz] < 0)
									cubeIndex += 16 if (data[ix, iy, iz+1] < 0)
									cubeIndex += 32 if (data[ix+1, iy, iz+1] < 0)
									cubeIndex += 64 if (data[ix+1, iy+1, iz+1] < 0)
									cubeIndex += 128 if (data[ix, iy+1, iz+1] < 0)
									
									next if (cubeIndex == 0) or (cubeIndex == 255)
									
									edges[0]  = points[ix, iy, iz][0]
									edges[1]  = points[ix+1, iy, iz][1]
									edges[2]  = points[ix, iy+1, iz][0]
									edges[3]  = points[ix, iy, iz][1]
									edges[4]  = points[ix, iy, iz+1][0]
									edges[5]  = points[ix+1, iy, iz+1][1]
									edges[6]  = points[ix, iy+1, iz+1][0]
									edges[7]  = points[ix, iy, iz+1][1]
									edges[8]  = points[ix, iy, iz][2]
									edges[9]  = points[ix+1, iy, iz][2]
									edges[10] = points[ix+1, iy+1, iz][2]
									edges[11] = points[ix, iy+1, iz][2]
									
									tris = MARCHING_CUBES[cubeIndex]
									
									0.step(12, 3) do |i|
										break if tris[i] == -1
										
										v1 = edges[tris[i]]
										v2 = edges[tris[i+1]]
										v3 = edges[tris[i+2]]
										
										next unless (v1 && v2 && v3)
										next if ((v1 == v2) || (v2 == v3) || (v3 == v1))
										
										mesh.add_polygon v1, v3, v2
									end
								end
							end
						else
							SurfaceGen.finishGenerate(entities, mesh, softEdges, boundary)
						end
					end
				end
			end
		end
	end
	
	def SurfaceGen.updateJS()
		selection = Sketchup::active_model.selection
		return unless ((selection.length == 1) and (selection[0].get_attribute(ATTR_DICTIONARY, "isGraph", false)))
		
		surfaceGen = selection[0]
		
		attr = surfaceGen.attribute_dictionary(ATTR_DICTIONARY)
		
		script = ""
		attr.each_pair { |key, value|
			if value.is_a? String then
				script += "attr['" + key + "']='" + value.gsub(/(['\\])/, '\\\1') + "'; "
			else
				script += "attr['" + key + "']=" + value.to_s + "; "
			end
		}
		script += "update(); "
		SurfaceGen::dialog.execute_script(script)
	end
	
	def SurfaceGen.saveJS(params)
		key, value = *params.split("=", 2)
		selection = Sketchup::active_model.selection
		return unless ((selection.length == 1) and (selection[0].get_attribute(ATTR_DICTIONARY, "isGraph", false)))
		selection[0].set_attribute(ATTR_DICTIONARY, key, eval(value))
	end
			
	def SurfaceGen.generateHeights(eqn, xsize, ysize, xoffset, xscale, yoffset, yscale, &proc)
		data = Array2D.new(xsize+1, ysize+1)
		eqnProc = eval("Proc.new {|x,y| " + eqn + "}")
	
		SurfaceGen.asyncLoop(0..ysize) do |iy, last|
			if (!last) then
				y = iy*yscale + yoffset
				(0..xsize).each do |ix|
					x = ix*xscale + xoffset
					
					begin
						# data[ix,iy] = eval(eqn).to_f
						data[ix,iy] = eqnProc.call(x,y)
						
						if (data[ix, iy].nan? or data[ix, iy].infinite?) then
							data[ix, iy] = nil
						end
					rescue Exception => exc
						data[ix,iy] = nil
					end
				end
			else
				proc.call(data)
			end
		end
	end
	
	def SurfaceGen.generateVolume(eqn, xsize, ysize, zsize, xoffset, xscale, yoffset, yscale, zoffset, zscale, &proc)
		data = Array3D.new(xsize, ysize, zsize)
		eqnProc = eval("Proc.new {|x,y,z| " + eqn + "}")
	
		SurfaceGen.asyncLoop(0..zsize) do |iz, last|
			if (!last) then
				z = iz*zscale + zoffset
				(0..ysize).each do |iy|
					y = iy*yscale + yoffset
					(0..xsize).each do |ix|
						x = ix*xscale + xoffset
						
						begin
							# data[ix,iy,iz] = eval(eqn).to_f
							data[ix,iy,iz] = eqnProc.call(x,y,z)
							
						rescue Exception => exc
							data[ix,iy,iz] = 0
						end
					end
				end
			else
				proc.call(data)
			end
		end
	end

	# Add menu items
	if not file_loaded? "surfacegen.rb" then
		require(File.join(SURFACEGEN_DIR, "array_nd.rb"))
		require(File.join(SURFACEGEN_DIR, "marching_cubes.rb"))
		require(File.join(SURFACEGEN_DIR, "surfacegen_tool.rb"))
		require(File.join(SURFACEGEN_DIR, "surfacegen_observers.rb"))
		
		@myTool = SurfaceGenTool.new
		
		toolbar = UI::Toolbar.new "SurfaceGen"
		cmd = UI::Command.new("Add SurfaceGen Graph") { Sketchup.active_model.select_tool @myTool }
		cmd.small_icon = File.join(PLUGINS_DIR, SURFACEGEN_DIR, "surfacegen_small.png")
		cmd.large_icon = File.join(PLUGINS_DIR, SURFACEGEN_DIR, "surfacegen.png")
		cmd.tooltip = "Add SurfaceGen Graph"
		cmd.status_bar_text = "Create a 3D surface generated from a mathematical formula."
		cmd.set_validation_proc {
			if @myTool.active?
				MF_CHECKED
			else
				MF_ENABLED
			end
		}
		
		toolbar.add_item cmd
		toolbar.restore
		
		Sketchup.active_model.selection.add_observer(SurfaceGenSelectionObserver.new)
		
		keys = {
			:dialog_title => "SurfaceGen Options",
			:scrollable => false,
			:preferences_key => 'SurfaceGenOptions',
			:height => WINDOW_HEIGHT,
			:width => WINDOW_WIDTH,
			:left => 200,
			:top => 200,
			:resizable => false,
			:mac_only_use_nswindow => true
		}
		begin
			@dialog = UI::WebDialog.new(keys);
		rescue
			@dialog = UI::WebDialog.new("SurfaceGen Options", false, 'SurfaceGenOptions', 200, 200, 0);
			@dialog.set_size WINDOW_WIDTH, WINDOW_HEIGHT
		end

		@dialog.set_file(File.join(PLUGINS_DIR, SURFACEGEN_DIR, "SurfaceGenOptions.htm"))
		@dialog.add_action_callback("save") {|dialog, params|
			data = ""
			0.step(params.length-2, 2) { |i|
				data += params[i..(i+1)].to_i(16).chr
			}
		
			SurfaceGen.saveJS(data)
		}
		@dialog.add_action_callback("generate") {|dialog, params| SurfaceGen.generate }
		@dialog.add_action_callback("refresh") {|dialog, params| SurfaceGen.updateJS }


		file_loaded("surfacegen.rb")
	end
end
